
from multiprocessing import Pool

import loguru


def task(a):
    loguru.logger.warning(f"{a}asfasfasfasfasf")
    loguru.logger.warning(f"{a}asfasfasfasfasf")
    loguru.logger.warning(f"{a}asfasfasfasfasf")
    loguru.logger.warning(f"{a}asfasfasfasfasf")

def main():
    with Pool(5) as pl:
        pl.map(task,[i for i in range(0,19)])


if __name__ == '__main__':
    main()


from __future__ import annotations

import asyncio
import datetime
import os
import uuid
from dataclasses import dataclass, field
from functools import partial
from inspect import iscoroutine, iscoroutinefunction, isawaitable
from time import time_ns
from typing import Dict, Callable, Coroutine, Any, TypeVar, Generic

import loguru
from apscheduler.events import EVENT_JOB_SUBMITTED, JobSubmissionEvent, JobExecutionEvent, \
    EVENT_JOB_ERROR, EVENT_JOB_EXECUTED
from apscheduler.executors.base import BaseExecutor
from apscheduler.executors.pool import ProcessPoolExecutor
from apscheduler.job import Job
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.base import BaseTrigger
from pytz import utc

from web_foundation.kernel.abstract.container import GenericDepContainer
from web_foundation.kernel.worker import Worker

BackgroundTask = TypeVar("BackgroundTask", Callable[..., Coroutine[Any, Any, Any]], partial)


@dataclass
class Task:
    id: str
    status: str
    name: str
    scheduled_trigger: BaseTrigger
    scheduled_job: Job | None = field(default=None)
    on_error_callback: TaskErrorCallback = field(default=None)
    next_call_time: datetime.datetime | None = field(default=None)
    execution_time: float = field(default=0)
    done_time: float = field(default=0)
    call_time: float = field(default=0)
    call_counter: int = field(default=0)
    create_time: datetime.datetime | None = field(default_factory=datetime.datetime.now)
    error: BaseException | None = field(default=None)


TaskErrorCallback = Callable[[JobExecutionEvent, Task], Coroutine[Any, Any, None]]

from web_foundation.kernel import IMessage


class TaskIMessage(IMessage):
    message_type = "background_task"
    task: BackgroundTask
    args: Any | None
    kwargs: Any | None
    trigger: BaseTrigger
    add_job_kw: Dict
    on_error_callback: TaskErrorCallback

    def __init__(self, task: BackgroundTask,
                 trigger=None,
                 args: Any = None,
                 kwargs: Any | None = None,
                 add_job_kw: Dict = None,
                 on_error_callback: TaskErrorCallback = None):
        super().__init__()
        self.task = task
        self.args = args
        self.kwargs = kwargs
        self.add_job_kw = add_job_kw if add_job_kw else {}
        self.trigger = trigger
        self.on_error_callback = on_error_callback


class TaskExecutor(Worker, Generic[GenericDepContainer]):
    _tasks: Dict[str, Task]
    _scheduler: BackgroundScheduler

    def __init__(self, **kwargs):
        self._tasks = {}
        super().__init__(**kwargs)

    def _configure(self, *args, executors: Dict[str, BaseExecutor] = None, job_defaults=None, **kwargs):
        executor = {
            'default': ProcessPoolExecutor(os.cpu_count())
        }
        job_default = {
            "misfire_grace_time": 1,
            'coalesce': False,
            'max_instances': 10
        }
        self._scheduler = BackgroundScheduler(jobstores={"default": MemoryJobStore()},
                                              executors=executors if executors else executor,
                                              job_defaults=job_defaults if job_defaults else job_default,
                                              timezone=utc)
        self._add_listeners()

    async def _run(self):
        self.channel.add_event_listener(TaskIMessage, self._task_message_handler)
        self._scheduler.start()
        await self.channel.listen_consume()

    def _get_callable(self, fnc: BackgroundTask):
        caller = fnc
        if iscoroutine(fnc) or isawaitable(fnc):
            def __call(*args, **kwargs):
                loop = asyncio.get_running_loop()
                loop.run_until_complete(fnc)

            caller = __call

        elif iscoroutinefunction(fnc):
            def __call(*args, **kwargs):
                loop = asyncio.get_running_loop()
                loop.run_until_complete(fnc(*args, **kwargs))

            caller = __call

        return caller

    async def add_task(self, struct: TaskIMessage):
        await self._task_message_handler(struct)

    async def _task_message_handler(self, message: TaskIMessage):
        message.task = self._get_callable(message.task)
        arc_task = Task(id=str(uuid.uuid4()),
                        name=message.task.__name__,
                        status="new",
                        scheduled_trigger=message.trigger,
                        on_error_callback=message.on_error_callback)
        arc_task.scheduled_job = self._scheduler.add_job(message.task,
                                                         name=arc_task.name,
                                                         id=arc_task.id,
                                                         trigger=message.trigger,
                                                         args=message.args,
                                                         kwargs=message.kwargs,
                                                         **message.add_job_kw)
        arc_task.next_call_time = arc_task.scheduled_job.next_run_time
        self._tasks.update({arc_task.id: arc_task})

    def _get_task(self, id: str):
        return self._tasks.get(id)

    def _add_listeners(self):
        def _on_job_submitted(event: JobSubmissionEvent):
            nonlocal self
            task = self._get_task(event.job_id)
            task.call_time = time_ns()
            task.call_counter += 1
            task.status = "called"

        def _on_job_exec(event: JobExecutionEvent):
            nonlocal self
            task = self._get_task(event.job_id)
            task.done_time = time_ns()
            task.execution_time = task.done_time - task.call_time
            task.status = "done"
            task.next_call_time = task.scheduled_job.next_run_time
            if event.exception:
                task.error = event.exception
                task.status = 'error'
                if task.on_error_callback:
                    asyncio.create_task(task.on_error_callback(event, task))

        self._scheduler.add_listener(_on_job_submitted, EVENT_JOB_SUBMITTED)
        self._scheduler.add_listener(_on_job_exec, EVENT_JOB_EXECUTED | EVENT_JOB_ERROR)

    async def after_worker_fork(self):
        pass

    def post_configure(self):
        pass

    async def close(self, *args, **kwargs):
        self._scheduler.shutdown()



import asyncio
from asyncio import Future
from typing import Dict, List, Any

import loguru

from web_foundation import settings
from web_foundation.kernel import IChannel, GenericIMessage, EventListener
from web_foundation.kernel.worker import GenWorker


class IDispatcher:
    channels: Dict[str, IChannel]
    events_listeners: Dict[str, EventListener]
    _msg_global_index: int

    def __init__(self):
        self._msg_global_index = 0
        self.channels = {}
        self.events_listeners = {}

    def reg_worker(self, worker: GenWorker):
        self.channels.update({worker.name: worker.channel})

    def add_event_listener(self, event_type: str, callback: EventListener):
        self.events_listeners[event_type] = callback

    async def track_event(self, msg: GenericIMessage):
        listener = self.events_listeners.get(msg.message_type)
        if listener:
            await listener(msg)

    async def on_channel_sent(self, msg: GenericIMessage):
        self._msg_global_index += 1
        msg.index = self._msg_global_index
        asyncio.create_task(self.track_event(msg))
        if msg.destination != "__dispatcher__":
            asyncio.create_task(self.broadcast(msg))

    async def broadcast(self, msg: GenericIMessage):
        for worker_name, ch in self.channels.items():
            if msg.sender == worker_name:
                continue
            if msg.destination in ["__all__", worker_name]:
                if settings.DEBUG:
                    loguru.logger.debug(f"Sent {msg} to {ch.worker_name}")
                await ch.consume_pipe.write(msg)

    def preform(self) -> List[Future[Any]]:
        tasks = []
        for channel in self.channels.values():
            tasks.append(asyncio.ensure_future(
                channel.listen_produce(
                    self.on_channel_sent
                )))
        return tasks



import asyncio
import json as py_json
from typing import List, Type, Coroutine, Callable, Any

from sanic.exceptions import NotFound

from web_foundation.environment.resources.database.model_loader import ModelLoader
from web_foundation.environment.resources.database.models import AbstractDbModel, GenericDbType
from web_foundation.environment.workers.web.ext.request_handler import InputContext

ExternalCallback = Callable[
    [AbstractDbModel | List[AbstractDbModel] | None], Coroutine[Any, Any, None]]



async def exec_access(inc: InputContext,
                      model: Type[AbstractDbModel],
                      fetch_fields: List[str] = None,
                      middleware: Type[ModelLoader] = ModelLoader,
                      external_callback: ExternalCallback = None
                      ):
    """
    Access to models
    """
    match inc.request.method:
        case "GET":
            kwargs = inc.r_kwargs
            entity_id = kwargs.pop("entity_id", None)
            if entity_id is None:
                limit = inc.request.args.get("limit")
                offset = inc.request.args.get("offset")
                limit = int(limit) if limit and limit.isdigit() else 100
                offset = int(offset) if offset and offset.isdigit() else None
                # order_by = inc.request.args.get("order_by")  # for openapi. Variable forwarded to read_all in kwargs
                for ar, val in inc.request.args.items():
                    if ar in ["limit", "offset"]:
                        continue
                    val = val[0]
                    if val.lower() == 'false':
                        val = False
                    elif val.lower() == 'true':
                        val = True
                    elif val.isdigit():
                        val = int(val)
                    elif "." in val and val.replace('.', '').isdigit():
                        val = float(val)
                    elif val.startswith("["):
                        val = py_json.loads(val)
                    kwargs[ar] = val

                retrieved_all, total = await middleware.read_all(model, inc.identity, limit, offset,
                                                                 fetch_fields=fetch_fields, **kwargs)
                if external_callback:
                    asyncio.create_task(external_callback(retrieved_all))
                res = [await model.values_dict() for model in retrieved_all]
                return res, total
            retrieved = await middleware.read(model, entity_id, inc.identity, fetch_fields=fetch_fields, **kwargs)
            if not retrieved:
                raise NotFound()

        case "POST":
            retrieved = await middleware.create(model, inc.identity, inc.dto, **inc.r_kwargs)

        case "PATCH":
            entity_id = inc.r_kwargs.pop("entity_id")
            retrieved = await middleware.update(model, entity_id, inc.identity, inc.dto, **inc.r_kwargs)

        case "DELETE":
            entity_id = inc.r_kwargs.pop("entity_id")
            retrieved = await middleware.delete(model, entity_id, inc.identity, **inc.r_kwargs)

    if external_callback:
        asyncio.create_task(external_callback(retrieved))
    if isinstance(retrieved, list):
        return [await i.values_dict() for i in retrieved]
    return await retrieved.values_dict()



from __future__ import annotations

from pathlib import Path
from typing import Generic, TypeVar

from pydantic import BaseSettings

from web_foundation.environment.resources.file_repo.repo import FileRepository
from web_foundation.environment.resources.file_repo.system_repo import SystemFileRepository
from web_foundation.environment.resources.stores.memory import InMemoryDictStore
from web_foundation.environment.resources.stores.store import AppStore
from web_foundation.environment.services.service import Service
from web_foundation.kernel.abstract.container import AbstractDependencyContainer, Dependency
from web_foundation.kernel.abstract.resource import Resource
from web_foundation.kernel.configuration import GenericConfig
from web_foundation.utils.helpers import in_obj_subclasses, in_obj_classes, in_obj_cls_items


class AppDependencyContainer(AbstractDependencyContainer, Generic[GenericConfig]):
    app_config: Dependency[GenericConfig] = Dependency(instance_of=BaseSettings)
    store: Dependency[AppStore] = Dependency(instance_of=AppStore, default=InMemoryDictStore())
    file_repository: Dependency[FileRepository] = Dependency(instance_of=FileRepository,
                                                             default=SystemFileRepository(root=Path("applied_files")))

    async def init_resources(self):
        for resource, name in in_obj_subclasses(self, Resource):
            await resource.init(self.app_config())

    async def shutdown_resources(self):
        for resource, name in in_obj_subclasses(self, Resource):
            await getattr(self, name).shutdown()

    def services(self):
        return set(in_obj_subclasses(self, Service))


GenericAppDepContainer = TypeVar("GenericAppDepContainer", bound=AppDependencyContainer)



import importlib
import os
from typing import List, Generic

from loguru import logger
from tortoise import Tortoise, BaseDBAsyncClient, connections

from web_foundation.kernel.abstract.resource import Resource
from web_foundation.kernel.configuration import DbConfig
from web_foundation.kernel.configuration import GenericConfig


class DatabaseResource(Resource, Generic[GenericConfig]):
    connection: BaseDBAsyncClient | None
    _app_name: str
    _modules: List[str]
    _engine: str
    db_conf: DbConfig

    def __init__(self, modules: List[str], engine: str = 'tortoise.backends.asyncpg'):
        self._engine = engine
        self._modules = modules

    def _get_connection_setting(self) -> dict:
        if self.db_conf.with_migrations:
            self._modules.append("aerich.models")
        to_discover = [importlib.import_module(i) for i in self._modules]
        return {
            'connections': {
                # Dict format for connection
                f'{self._app_name}_default': {
                    'engine': self._engine,
                    'credentials': {
                        'host': self.db_conf.host,
                        'port': self.db_conf.port,
                        'user': self.db_conf.user,
                        'password': self.db_conf.password,
                        'database': self.db_conf.database,
                        'schema': self.db_conf.db_schema,
                        'minsize': 1,
                        'maxsize': 5,
                    }
                }
            },
            'apps': {
                f'{self._app_name}': {
                    'models': to_discover,
                    'default_connection': f'{self._app_name}_default',
                }
            },
            'use_tz': False,
            'timezone': 'UTC'
        }

    async def fill_db_data(self):
        pass

    async def init(self, config: GenericConfig, **kwargs) -> BaseDBAsyncClient:
        self._app_name = config.app_name
        self.db_conf = config.database
        self.connection = None
        await Tortoise.init(config=self._get_connection_setting())
        await self.configure_db()
        return Tortoise.get_connection(f'{self._app_name}_default')

    async def shutdown(self) -> None:
        await connections.close_all()

    @property
    def conn(self) -> BaseDBAsyncClient:
        if not self.connection:
            self.connection = Tortoise.get_connection(connection_name=f'{self._app_name}_default')
        return self.connection

    async def _migrations(self, schema_exists: bool, command):
        path_exists = os.path.exists(os.path.join(os.getcwd(), self.db_conf.migrations_path)) and os.listdir(
            os.path.join(os.getcwd(), self.db_conf.migrations_path))

        if not path_exists and not schema_exists:
            await self.create_schema(False)
            await command.init()
            await command.init_db(safe=True)
        elif not schema_exists:
            await self.create_schema(False)
            await command.init()
            await command.upgrade()
        elif not path_exists:
            await command.init()
            await command.init_db(safe=True)  # TODO check - need to drop aerich table in db?
        await command.init()
        logger.info(f"Apply migrations from {self.db_conf.migrations_path}")
        await command.migrate()
        await command.upgrade()

    async def configure_db(self):
        scheme_name = self.db_conf.db_schema
        row_count, rows = await self.conn.execute_query(
            "SELECT schema_name FROM information_schema.schemata WHERE schema_name = $1", [scheme_name])
        schema_exists = True if row_count else False
        if self.db_conf.with_migrations:
            if not self.db_conf.migrations_path:
                raise ValueError("Migration Path not set can't migrate")
            try:
                from aerich import Command
            except ImportError:
                raise ImportError("To use migration need to install aerich, (web-foundation[aerich])")
            command = Command(tortoise_config=self._get_connection_setting(), app=self._app_name,
                              location=self.db_conf.migrations_path)
            await self._migrations(schema_exists, command)
        if not schema_exists:
            await self.create_schema()
            await self.fill_db_data()

    async def create_schema(self, generate_schemas: bool = True):
        await self.conn.execute_script(f"CREATE SCHEMA IF NOT EXISTS {self.db_conf.db_schema};")
        if generate_schemas:
            await Tortoise.generate_schemas()



from copy import copy
from typing import Any, Union, Dict, List, Type

from orjson import dumps
from orjson import loads

from web_foundation.environment.events.store import StoreUpdateEvent
from web_foundation.kernel import IChannel

JSON = Union[Dict[str, Any], List[Any], int, str, float, bool, Type[None]]
TypeJSON = Union[Dict[str, 'JSON'], List['JSON'], int, str, float, bool, Type[None]]


class AppStore:
    _channel: IChannel = None
    need_sync: bool = False

    async def on_store_update(self, event: StoreUpdateEvent):
        if not event.obj:
            event.value = self._load(event.value)
        await self.set_item(event.key, event.value, send_event=False)

    async def get_item(self, key: str) -> Any:
        raise NotImplementedError

    async def _set_item(self, key: str, value: TypeJSON):
        raise NotImplementedError

    async def set_item(self, key: str, value: Any, obj: bool = False, send_event: bool = True):
        await self._set_item(key, value)
        if self.need_sync and send_event:
            data = value if obj else self._serialize(value)
            await self._channel.produce(StoreUpdateEvent(key, copy(data), obj))

    async def get_all(self):
        raise NotImplementedError

    async def size(self):
        raise NotImplementedError

    def _load(self, value: str):
        return loads(value)

    def _serialize(self, value: TypeJSON):
        return dumps(value)

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, channel: IChannel):
        self._channel = channel
        if self.need_sync:
            self._channel.add_event_listener(StoreUpdateEvent, self.on_store_update)



from web_foundation.environment.services.service import Service


class ApiAddonsService(Service):

    async def add_new_middleware(self, filename: str):
        await self._worker.mdw_manager.add_new_middleware(filename)

    async def delete_middleware(self, filename: str):
        await self._worker.mdw_manager.delete_middleware(filename)



from __future__ import annotations

import asyncio
import json
from asyncio import Future
from pathlib import Path
from typing import List, Type, Generic, Dict, Any

from web_foundation import settings
from web_foundation.environment.container import GenericAppDepContainer
from web_foundation.kernel.abstract.app import AbstractApp
from web_foundation.kernel.abstract.container import GenericDepContainer
from web_foundation.kernel.configuration import GenericConfig
from web_foundation.environment.metrics.dispatcher import MetricsDispatcher
from web_foundation.environment.resources.file_repo.repo import FileRepository
from web_foundation.environment.resources.stores.store import AppStore
from web_foundation.environment.services.service import Service
from web_foundation.kernel.messaging.dispatcher import IDispatcher
from web_foundation.kernel.worker import GenWorker, Worker


class App(AbstractApp, Generic[GenericAppDepContainer]):
    name: str
    workers: Dict[str, Worker]
    services: Dict[str, Service]
    dispatcher: IDispatcher
    container: GenericAppDepContainer

    # debug: bool  # type: ignore

    def __init__(self, container: GenericAppDepContainer):
        super().__init__(container)
        self.name = container.app_config().app_name
        self.dispatcher = MetricsDispatcher() if settings.METRICS_ENABLE else IDispatcher()

    async def _before_app_run(self):
        await self.container.init_resources()
        await self.before_app_run()
        await self.container.shutdown_resources()

    async def before_app_run(self):
        pass

    def _add_worker(self, worker: GenWorker):
        if not worker.configured:
            raise RuntimeError("Worker not configured")
        worker.store = self.container.store()
        worker.repo = self.container.file_repository()
        worker.app_name = self.name
        worker.container = self.container
        self.dispatcher.reg_worker(worker)
        worker.post_configure()
        self.workers.update({worker.name: worker})

    def add_worker(self, worker: GenWorker | List[GenWorker]):
        """
        Set isolate to environment and set isolate debug and name
        :param worker: isolate to apply
        :return: None
        """
        if isinstance(worker, list):
            for w in worker:
                self._add_worker(w)
        else:
            self._add_worker(worker)

    def add_service(self, service: Service):
        self.services.update({service.__class__.__name__: service})

    def find_worker_by_pid(self, pid: int):
        for worker in self.workers.values():
            if worker.pid == pid:
                return worker

    def perform(self) -> List[Future[Any]]:
        """
        Call performs from isolates and return all isolates Futures
        :return: None
        """
        return [worker.perform() for worker in self.workers.values()]

    async def run(self):
        await self._before_app_run()
        """
        Func to run environment manually (without Runner)
        :return: None
        """
        return await asyncio.wait(self.perform() + self.dispatcher.preform())
